/*
 * Copyright (c) 2009-2011 CNRS/LAAS
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 ** rmp400MotionTaskCodels.c
 **
 ** Codels called by execution task rmp400MotionTask
 **
 ** Author: Matthieu Herrb
 ** Date: April 2009
 **
 **/
#include <sys/types.h>
#include <errno.h>
#include <math.h>
#include <pthread.h>
#include <stdlib.h>

#include <portLib.h>
#include <h2mathLib.h>
#include "h2timeLib.h"

#include <rmp/rmpLib.h>

#include "rmp400Const.h"
#include "server/rmp400Header.h"

#ifdef WITH_FELIB
#include "fe/ftdi-emergency.h"
#endif

#include "odoProba.h"
#include "joystick.h"

#ifdef GYRO
#include <gyroLib.h>
#if (GYRO==gyroKVH1000Task)
#include <gyroKVH1000TaskLib.h>
#endif
#endif

static struct RMP_DEV_STR **rmp;
static POSTER_ID pomPosterId = NULL;
static JOYSTICK_STR *joy = NULL;
#ifdef WITH_FELIB
static FE_STR *fe = NULL;
#endif

extern void *rmp400ReadTask(void *);
extern STATUS rmp400Odo3d(int *report, double period);

static STATUS rmp400DataUpdate(RMP400_STATUS_STR *);
static STATUS rmp400VelocitySet(double, double, int *);
static STATUS initPomPosters(void);
static STATUS updatePomPosters(void);
static STATUS rmp400VelocityGet(void);
static void getJoystickSpeeds(double *v, double *w, double *avmax, double *awmax);


/*------------------------------------------------------------------------
 *
 * initOdoAndAsserv  --  Initialization codel (fIDS, ...)
 *
 * Description:
 *
 * Returns:    OK or ERROR
 */

STATUS
initOdoAndAsserv(int *report)
{
	GENPOS_CART_STATE *configId;
	GENPOS_CART_REF *refId;
	GENPOS_CART_CMD_PARAM *cmdId;
	RMP400_KINEMATICS_STR *kinematics;
	RMP400_GYRO *gyro;

	/* Kinematics */
	kinematics = &rmp400DataStrId->kinematics;
	kinematics->leftWheelRadius = RMP400_LEFT_WHEEL_RADIUS;
	kinematics->rightWheelRadius = RMP400_RIGHT_WHEEL_RADIUS;
	kinematics->axisWidth = RMP400_AXIS_WIDTH;

	/* configuration */
	configId = &rmp400DataStrId->robot;
	configId->xRob = 0.;
	configId->yRob = 0.;
	configId->theta = 0.;
	configId->xRef = RMP400_DIST_TO_CONTROLLED_POINT;
	configId->yRef = 0;
	configId->v = 0;
	configId->w = 0;

	/* Ref Data */
	refId = &rmp400DataStrId->ref;
	refId->dataType = GENPOS_POS_AND_SPEED_DATA;
	refId->backFlag = GENPOS_FORWARD_MOTION;
	refId->x = configId->xRef;
	refId->y = configId->yRef;
	refId->theta = configId->theta;
	refId->v = 0.;
	refId->w = 0.;

	refId->vmax = RMP400_VMAX;
	refId->wmax = RMP400_WMAX;
	refId->linAccelMax = RMP400_ACCEL_LIN_MAX;
	refId->angAccelMax = RMP400_ACCEL_ANG_MAX;

	/* Control parameters */
	cmdId = &rmp400DataStrId->cmd;
	cmdId->distPoint = RMP400_DIST_TO_CONTROLLED_POINT;
	cmdId->KpLongit = RMP400_KP_LONGIT;
	cmdId->KiLongit = RMP400_KI_LONGIT;
	cmdId->KpTransv = RMP400_KP_TRANSV;
	cmdId->KiTransv = RMP400_KI_TRANSV;

	cmdId->KpAng = RMP400_KP_ANG;
	cmdId->KiAng = RMP400_KI_ANG;

	cmdId->longitMaxGap = RMP400_LONGIT_MAX_GAP;
	cmdId->transvMaxGap = RMP400_TRANSV_MAX_GAP;
	cmdId->angMaxGap    = RMP400_ANG_MAX_GAP;

	cmdId->longitErrorMax = RMP400_LONGIT_ERROR_MAX;
	cmdId->transvErrorMax = RMP400_TRANSV_ERROR_MAX;
	cmdId->angErrorMax = RMP400_ANG_ERROR_MAX;

	cmdId->longitMaxErrSum = RMP400_LONGIT_MAX_ERR_SUM;
	cmdId->transvMaxErrSum = RMP400_TRANSV_MAX_ERR_SUM;
	cmdId->angMaxErrSum = RMP400_ANG_MAX_ERR_SUM;

	/* gyro */
	gyro = &rmp400DataStrId->gyro;
	gyro->initDone = FALSE;
	gyro->params.mode = RMP400_GYRO_ON;
#ifdef GYRO
	gyro->gyroOn = FALSE;
#endif
	gyro->gyroToRobotOffset = 0.0;
	gyro->gyroTheta = 0.0;

	/* Initialize Pom */
	initPomPosters();

	return OK;
}

double convertBattery(double voltage)
{
	int i;
	double calib[14][2] = {
		{   0., 0.00 },
		{   0., 2.90 },
		{   5., 3.20 },
		{  10., 3.50 },
		{  20., 3.60 },
		{  30., 3.65 },
		{  40., 3.69 },
		{  50., 3.73 },
		{  60., 3.77 },
		{  70., 3.82 },
		{  80., 3.87 },
		{  90., 4.00 },
		{ 100., 4.10 },
		{ 100., 9.99 }};
	double element = voltage / 20;
	
	for(i = 1; i < 14-1; ++i)
	{
		if (element < calib[i][1])
			return calib[i-1][0] + (calib[i][0]-calib[i-1][0]) * (element-calib[i-1][1])/(calib[i][1]-calib[i-1][1]);
	}
	return 100.;
}


/*----------------------------------------------------------------------*/
/*
 * rmp400DataUpdate - copy data from 2 motors block to the IDS
 */
static STATUS
rmp400DataUpdate(RMP400_STATUS_STR *status)
{
	static int prevFrame[2] = {0, 0};
	static int noData[2] = {0, 0};
	struct RMP_DATA_STR *data;
	struct RMP400_DATA_STR *dest;
	int i;

	for (i = 0; i < 2; i++) {
		data = rmpGetData(rmp[i]);
		if (data == NULL) {
			fprintf(stderr, "NULL state str %d\n", i);
			continue;
		}
#if 0
		if (data->servo_frames == prevFrame[i]) {
			fprintf(stderr,
				"Data not refreshed by rmp %d: %d\n",
				i, data->servo_frames);
			return ERROR;
		}
#endif
		prevFrame[i] = data->servo_frames;
		dest = &status->rs_data[i];

		pthread_mutex_lock(rmpGetMutex(rmp[i]));

		if (data->operational_mode != 1) {
			pthread_mutex_unlock(rmpGetMutex(rmp[i]));
			dest->operational_mode = data->operational_mode;
			noData[i]++;
			break;
		}
		noData[i] = 0;
		dest->pitch_angle = RMP400_CONVERT_PITCH(data->pitch_angle);
		dest->pitch_rate = RMP400_CONVERT_PITCH(data->pitch_rate);
		dest->roll_angle = RMP400_CONVERT_ROLL(data->roll_angle);
		dest->lw_velocity = RMP400_CONVERT_WHEEL_SPEED(data->lw_velocity);
		dest->rw_velocity = RMP400_CONVERT_WHEEL_SPEED(data->rw_velocity);
		dest->yaw_rate = RMP400_CONVERT_YAW(data->yaw_rate);
		dest->servo_frames = data->servo_frames;
		dest->integrated_left_wheel =
		   RMP400_CONVERT_INTEGRATED_POS(data->integrated_left_wheel);
		dest->integrated_right_wheel =
		   RMP400_CONVERT_INTEGRATED_POS(data->integrated_right_wheel);
		dest->integrated_fore_aft =
		   RMP400_CONVERT_INTEGRATED_POS(data->integrated_fore_aft);
		dest->integrated_yaw =
		   RMP400_CONVERT_INTEGRATED_TURN(data->integrated_yaw);
		dest->left_torque = RMP400_CONVERT_TORQUE(data->left_torque);
		dest->right_torque = RMP400_CONVERT_TORQUE(data->right_torque);
		dest->operational_mode = data->operational_mode;
		dest->controller_gain_schedule =
		    data->controller_gain_schedule;
		dest->ui_voltage = RMP400_CONVERT_UI_VOLTAGE(data->ui_voltage);
		dest->powerbase_voltage =
		    RMP400_CONVERT_POWERBASE_VOLTAGE(data->powerbase_voltage);
		dest->battery_charge = convertBattery(dest->powerbase_voltage);
		dest->velocity_command = data->velocity_command; /* ?? */
		dest->turn_command = data->turn_command; /* ?? */

		pthread_mutex_unlock(rmpGetMutex(rmp[i]));
	}
	/* If not data for several periods - set motors off */
	if (status->rs_mode != RMP400_MOTORS_OFF &&
	    (noData[0] > 10 || noData[1] > 10)) {
		printf("No data - Motors OFF %d %d?\n",
		       status->rs_data[0].operational_mode,
		       status->rs_data[1].operational_mode);

		status->rs_mode = RMP400_MOTORS_OFF;
	}
#ifdef WITH_FELIB
	/* Check emergency stop */
	fe_get_status(fe);
	/* ignore pause if motors are off */
	if ((fe_pins(fe) & FE_PAUSE) != 0 &&
	    status->rs_mode != RMP400_EMERGENCY &&
	    status->rs_mode != RMP400_MOTORS_OFF) {
		printf("Emergency pause!\n");
		if (status->rs_mode != RMP400_MOTORS_OFF) 
			status->rs_mode = RMP400_EMERGENCY;
	}
#endif
	return OK;
}



/*------------------------------------------------------------------------
 *
 * odo  --  1st codel of permanent activity (called BEFORE other activities)
 *
 * Description: Reads all robots state and computes odometry
 *
 * Reports:      OK
 *              S_rmp400_EMERGENCY_STOP
 *
 * Returns:    OK or ERROR
 */

STATUS
odo(int *report)
{
	GENPOS_CART_REF *refId = &SDI_F->ref;
	GENPOS_CART_STATE *robot = &SDI_F->robot;
	RMP400_STATUS_STR *status = &SDI_F->status;
	RMP400_GYRO *gyro = &SDI_F->gyro;
	double direction;
#if DEBUG>=1
	static int count = 0;
#endif

	if (rmp == NULL || rmp[0] == NULL || rmp[1] == NULL)
		return OK;		/* not initialized yet */

	rmp400DataUpdate(status);

	/* Check motors status */
	if (status->rs_mode == RMP400_MOTORS_OFF &&
	    status->rs_data[0].operational_mode == 1 &&
	    status->rs_data[1].operational_mode == 1) {
		printf("Motors ON\n");
		status->rs_mode = RMP400_IDLE;
	}

	/* Read config */
	if (rmp400VelocityGet() != OK)
		return ERROR;
	if (refId->backFlag == GENPOS_FORWARD_MOTION)
		direction = 1;
	else if (refId->backFlag == GENPOS_BACKWARD_MOTION)
		direction = -1;
	else
		direction = 0;
	robot->xRef = robot->xRob
		+ direction*RMP400_DIST_TO_CONTROLLED_POINT*cos(robot->theta);
	robot->yRef = robot->yRob
		+ direction*RMP400_DIST_TO_CONTROLLED_POINT*sin(robot->theta);

	odoProba(&SDI_F->robot, &SDI_F->var,
	    SDI_F->kinematics.axisWidth, SDI_F->varParams.coeffLinAng,
		   EXEC_TASK_PERIOD(RMP400_MOTIONTASK_NUM));

	/* Gyro */
#ifdef GYRO
	if (gyro->initDone && gyro->currentMode != RMP400_GYRO_OFF) {
	  if (!(gyroReadCorrAngle(&gyro->gyroTheta))) {
			gyro->currentMode = RMP400_GYRO_OFF;
			gyro->initDone = FALSE;
		}
#if (GYRO==gyroKVH1000Task)
	  if (gyroRate(&gyro->gyroOmega, &gyro->gyroTemperature) != OK) {
			gyro->currentMode = RMP400_GYRO_OFF;
			gyro->initDone = FALSE;
		}
#endif

	}
#endif
	if (gyro->currentMode == RMP400_GYRO_OFF ||
	    (gyro->currentMode == RMP400_GYRO_ON_IF_MOTION
		&& fabs(robot->w) < 0.017
		&& fabs(robot->v) < 0.01)) {
		/* -- odo is reference */

		/* reset gyro offset to match odo */
		gyro->gyroToRobotOffset = angleLimit(robot->theta
		    - gyro->gyroTheta);
		/* switch back to odo */
		if (gyro->gyroOn) {
#if DEBUG>=1
			printf ("-> odo (%5.2f d) v %5.2f w%5.3f\n",
			    RAD_TO_DEG(robot->theta), robot->v, robot->w);
#endif
			gyro->gyroOn = FALSE;
		}
	} else {
		/* -- gyro is reference */
		robot->theta = angleLimit(gyro->gyroTheta +
		    gyro->gyroToRobotOffset);
		/* switch back to gyro */
		if (!gyro->gyroOn) {
#if DEBUG>=1
			printf ("-> gyro (%5.2f d) v %5.2f w%5.3f\n",
			    RAD_TO_DEG(robot->theta), robot->v, robot->w);
#endif
			gyro->gyroOn = TRUE;
		}
	}

	/*debug */
#if DEBUG>=1
	if (count==10) {
		printf ("MODE %s. gyro %5.3f + offset %5.3f = %5.3f =? %5.3f\n",
		    gyro->gyroOn ? "GYRO" : "ODO",
		    RAD_TO_DEG(gyro->gyroTheta),
		    RAD_TO_DEG(gyro->gyroToRobotOffset),
		    RAD_TO_DEG(angleLimit(gyro->gyroTheta +
			    gyro->gyroToRobotOffset)),
		    RAD_TO_DEG(robot->theta));
		count=0;
	}
	count++;
#endif
	if (SDI_F->odometryMode == RMP400_ODO_3D) {
	  rmp400Odo3d (report, EXEC_TASK_PERIOD(RMP400_MOTIONTASK_NUM));


	}
	return OK;
}

/*----------------------------------------------------------------------*/

/**
 ** Actual tracking code
 **/
static void
track(int *report, double *vRef, double *wRef)
{
	GENPOS_CART_REF *refId = &SDI_F->ref;
	GENPOS_CART_STATE *robot = &SDI_F->robot;
	GENPOS_TRACK *trackId = &SDI_F->trackStr;
	RMP400_STATUS_STR *status = &SDI_F->status;
	static double prevV, prevW;
	static BOOL prevDrifted = FALSE;
	BOOL drifted = FALSE;

	switch (trackId->trackMode) {
	case GENPOS_TRACK_POS:
		if (refId->dataType != GENPOS_POS_DATA) {
			fprintf(stderr, "wrong data type for TRACK_POS\n");
			*report = S_rmp400_BAD_REF;
			return;
		}
		drifted = contLawCartPositionControl(&rmp400DataStrId->cmd,
		    refId, robot, EXEC_TASK_PERIOD(RMP400_MOTIONTASK_NUM),
		    prevV, prevW, vRef, wRef);
		break;

	case GENPOS_TRACK_CONFIG:
		if (refId->dataType != GENPOS_POS_AND_SPEED_DATA) {
			fprintf(stderr, "wrong data type for TRACK_CONFIG\n");
			*report = S_rmp400_BAD_REF;
			return;
		}
		drifted = contLawCartConfigControl(&rmp400DataStrId->cmd,
		    refId, robot, EXEC_TASK_PERIOD(RMP400_MOTIONTASK_NUM),
		    prevV, prevW, vRef, wRef);
		break;

	case GENPOS_TRACK_SPEED:
		if (refId->dataType != GENPOS_SPEED_DATA) {
			fprintf(stderr, "wrong data type for TRACK_SPEED %d\n",
			    refId->dataType);
			*report = S_rmp400_BAD_REF;
			return;
		}
		*vRef = (refId->backFlag == GENPOS_FORWARD_MOTION ? refId->v :
		    -refId->v);
		*wRef = refId->w;
		break;
	default:
		*report = S_rmp400_BAD_REF;
		break;
	}
	if (drifted && !prevDrifted) {
		printf("odoAndAsserv: important drift!\n");
		printf("ref: %.2lf %.2lf %.2lf\n",
		       refId->x, refId->y, refId->theta);
		printf("pos: %.2lf %.2lf %.2lf\n",
		       robot->xRef, robot->yRef, robot->theta);
	}
	if (!drifted && prevDrifted)
		printf("odoAndAsserv: drift cancelled\n");
	prevDrifted = drifted;

	prevV = *vRef;
	prevW = *wRef;
}

/*------------------------------------------------------------------------
 *
 * asserv  --  2nd codel of permanent activity (called AFTER other activities)
 *
 * Description: Computes reference speeds and apply them
 *
 * Reports:      OK
 *              S_rmp400_EMERGENCY_STOP
 *
 * Returns:    OK or ERROR
 */

STATUS
asserv(int *report)
{
	GENPOS_CART_REF *refId = &SDI_F->ref;
	GENPOS_CART_STATE *robot = &SDI_F->robot;
	RMP400_STATUS_STR *status = &SDI_F->status;
	double vRef, wRef,dLongit, dTransv;

	if (rmp == NULL || rmp[0] == NULL || rmp[1] == NULL)
		return OK;		/* not initialized yet */

	*report = OK;
	switch (SDI_F->status.rs_mode) {
	case RMP400_MOTORS_OFF:
		/* No motion possible */
		return OK;

	case RMP400_EMERGENCY:
	case RMP400_IDLE:
		vRef = 0.0;
		wRef = 0.0;
		refId->x = robot->xRef;
		refId->y = robot->yRef;
		refId->theta = robot->theta;
		break;

	case RMP400_MANUAL:
		if (joystickPoll(joy) < 0) {
			*report = S_rmp400_JOYSTICK_ERROR;
			break;
		}
		refId->x = robot->xRef;
		refId->y = robot->yRef;
		refId->theta = robot->theta;
		getJoystickSpeeds(&vRef, &wRef, &refId->linAccelMax, &refId->angAccelMax);
		refId->v = vRef;
		refId->w = wRef;
		break;

	case RMP400_TRACK:
		track(report, &vRef, &wRef);
		break;

	default:
		*report = S_rmp400_BAD_REF;
		return ERROR;
	}
	if (*report != OK) {
		/* In case an error occured,
		   stop the robot and the tracking */
		SDI_F->status.rs_mode = RMP400_IDLE;
		SDI_F->trackStr.trackMode = GENPOS_NO_TRACKING;
		vRef = 0;
		wRef = 0;
	}

	/* Send commands to motors if ON */
	if (rmp400VelocitySet(vRef, wRef, report) != OK)
		if (*report == S_rmpLib_USB_WRITE_ERROR)
			/* USB doesn't accept commands ->
			   motors are probably off */
			status->rs_mode = RMP400_MOTORS_OFF;

	updatePomPosters();

	return OK;
}

/*------------------------------------------------------------------------
 * Init
 *
 * Description:
 *
 * Reports:      OK
 */

/* rmp400InitMain  -  codel EXEC of Init
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
rmp400InitStart(int *report)
{
	int i, n;
	char name[20];

	*report = OK;
#ifdef WITH_FELIB
	/* connect emergency stop */
	fe = fe_init();
	if (fe == NULL) {
		*report = S_rmp400_EMERGENCY_ERROR;
		return ETHER;
	}
#endif
	if ((n = rmpOpenAll(&rmp)) == -1) {
		*report = S_rmp400_RMPLIB_ERROR;
		return ETHER;
	}
	if (n != 2) {
		/* we expect 2 RMP devices */
		*report = S_rmp400_RMPLIB_ERROR;
		return ETHER;
	}
	printf("rmpOpenAll ok\n");
	/* start read tasks */
	errnoSet(0);
	for (i = 0; i < 2; i++) {
		snprintf(name, sizeof(name), "tReadTask%d", i);
		if (taskSpawn(name, 100, 0, 8192, rmp400ReadTask, rmp[i])
		    == ERROR) {
			fprintf(stderr, "taskSpawn %s error\n", name);
			*report =  errnoGet();
			break;
		}
		printf("%s spawned\n", name);
	}
	if (*report != OK) {
		fprintf(stderr, "rmp400InitMain: report 0x%x\n", *report);
		rmpClose(rmp[0]);
		rmpClose(rmp[1]);
		free(rmp);
	}
	for (i = 0; i < 2; i++)
		if (rmpPowerOn(rmp[i]) == ERROR) {
			*report = errnoGet();
			break;
		}
	return EXEC;
}

ACTIVITY_EVENT
rmp400InitMain(int *report)
{
	RMP400_STATUS_STR *status = &SDI_F->status;

	/* Check motors status */
	if (status->rs_data[0].operational_mode != 1 ||
	    status->rs_data[1].operational_mode != 1) {
		printf("Motors OFF\n");
		status->rs_mode = RMP400_MOTORS_OFF;
	} else {
		printf("Motors ON\n");
		status->rs_mode = RMP400_IDLE;
	}
	return ETHER;
}

/*------------------------------------------------------------------------
 * JoystickOn
 *
 * Description:
 *
 * Reports:      OK
 *              S_rmp400_BAD_REF
 *              S_rmp400_RMP_LIB_ERROR
 */

/* rmp400JoystickOnStart  -  codel START of JoystickOn
   Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
rmp400JoystickOnStart(int *report)
{
	int n;

	joy = joystickInit(RMP400_JOYSTICK_DEVICE_PATH);
	if (joy == NULL) {
		fprintf(stderr, "joystickInit(%s) failed: %s\n",
			RMP400_JOYSTICK_DEVICE_PATH, strerror(errno));
		*report = S_rmp400_JOYSTICK_ERROR;
		return ETHER;
	}
	n = joystickPoll(joy);
	memset(joy->buttons, 0, sizeof(joy->buttons));
	SDI_F->status.rs_mode = RMP400_MANUAL;
	return EXEC;
}

/* rmp400JoystickOnMain  -  codel EXEC of JoystickOn
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
rmp400JoystickOnMain(int *report)
{
	RMP400_STATUS_STR *status = &rmp400DataStrId->status;

	/* Check if mode changed */
	switch (status->rs_mode) {
	case RMP400_EMERGENCY:
		*report = S_rmp400_EMERGENCY_STOP;
		return END;
	case RMP400_MOTORS_OFF:
		*report = S_rmp400_MOTORS_OFF;
		return END;
	}
	/* Check for user abort (button 4) */
	if (joy->buttons[3]) {
		printf("Stop joystick\n");
		return rmp400JoystickOnInter(report);
	}
	return EXEC;
}

/* rmp400JoystickOnInter  -  codel INTER of JoystickOn
   Returns:  INTER ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
rmp400JoystickOnInter(int *report)
{
	GENPOS_CART_REF *refId;
	int n;

	n = joystickPoll(joy);

	refId = &rmp400DataStrId->ref;
	refId->linAccelMax = RMP400_ACCEL_LIN_MAX;
	refId->angAccelMax = RMP400_ACCEL_ANG_MAX;
	SDI_F->status.rs_mode = RMP400_IDLE;
	joystickEnd(joy);
	joy = NULL;
	return ETHER;
}

/*------------------------------------------------------------------------
 * SetPom
 *
 * Description:
 *
 * Reports:      OK
 *              S_rmp400_POSTER_NOT_FOUND
 */

/* setPom  -  codel EXEC of SetPom
   Returns:  EXEC END ETHER FAIL ZOMBIE */

ACTIVITY_EVENT
setPom(GENPOS_POSTER_NAME *posterName, int *report)
{
	POSTER_ID id;

	if (posterFind(posterName->name, &id) != OK) {
		logMsg("posterFind(%s) failed\n", posterName->name);
		*report = S_rmp400_POSTER_NOT_FOUND;
	} else
		pomPosterId = id;
	return ETHER;
}

/*----------------------------------------------------------------------*/
/*
 * Compute speed
 *
 * sets robot.v and robot.w in the SDI from motion in the motor structs.
 */


static STATUS
rmp400VelocityGet(void)
{
	double leftv, rightv;
	RMP400_DATA_STR *rmpData[2];
	GENPOS_CART_STATE *robot = &SDI_F->robot;
	RMP400_KINEMATICS_STR *kinematics = &SDI_F->kinematics;
	int i;

	for (i = 0; i < 2; i++)
		rmpData[i] = &SDI_F->status.rs_data[i];

	/* Mean angular speed of front and back wheels */
	leftv = (rmpData[0]->lw_velocity + rmpData[1]->lw_velocity) / 2.0;
	rightv = (rmpData[0]->rw_velocity + rmpData[1]->rw_velocity) / 2.0;
	/* printf("%.2lf %.2lf\n", leftv, rightv); */

	/* Angular speed of the robot */
#ifdef USE_INTERNAL_GYRO
	robot->w = (rmpData[0]->yaw_rate + rmpData[1]->yaw_rate) / 2.0;
#else
	robot->w = (leftv - rightv) / kinematics->axisWidth;
#endif
	/* linear speeds */
	robot->v = (leftv + rightv) / 2.0;

	return OK;
}

/*----------------------------------------------------------------------*/
/*
 * Send given linear and angular speeds to the hardware
 *
 */
static STATUS
rmp400VelocitySet(double v, double w, int *report)
{
	int i;

	/* Last safety check, in case uninitialized values are used */
	if (fabs(v) > RMP400_VMAX) {
		fprintf(stderr, "WARNING: v > VMAX: %lf\n", v);
		v = v > 0.0 ? RMP400_VMAX : -RMP400_VMAX;
	}
	if (fabs(w) > RMP400_WMAX) {
		fprintf(stderr, "WARNING: w > WMAX: %lf\n", w);
		w = w > 0.0 ? RMP400_WMAX : -RMP400_WMAX;
	}

	/* Send the same speed to front and back axis */
	for (i = 0; i < 2; i++) {
		if (rmpSendSpeed(rmp[i], RMP400_CONVERT_TO_RMP_SPEED(v),
				 RMP400_CONVERT_TO_RMP_ANGLE(w)) != OK) {
			*report = errnoGet();
			return ERROR;
		}
	}
	return OK;
}

/*----------------------------------------------------------------------*/
/*
 * Initialize POM
 */
static STATUS
initPomPosters(void)
{
	POM_ME_POS *mePos = &rmp400DataStrId->mePos;

	mePos->kind = POM_ME_ABSOLUTE;
	mePos->date1 = -1;
	mePos->confidence = 0.2;	// XXX used?
	mePos->main.var.voo     = RMP400_VAR_TT;
	mePos->main.var.vpp     = RMP400_VAR_TT;
	mePos->main.var.vrr     = RMP400_VAR_TT;
	mePos->main.var.vxx     = RMP400_VAR_XX;
	mePos->main.var.vyy     = RMP400_VAR_YY;
	mePos->main.var.vzz     = RMP400_VAR_XX;

	return OK;
}

/*----------------------------------------------------------------------*/

/*
 * Update POM poster
 */
static STATUS
updatePomPosters(void)
{
	POM_ME_POS *mePos = &SDI_F->mePos;
	GENPOS_CART_STATE *configId;
	RMP400_CART_3DSTATE *config3dId;
	POM_POS pomPos;

	if (pomPosterId != NULL) {
		if(rmp400POM_POSPosterRead(pomPosterId, &pomPos) == OK) {
			mePos->date1 = pomPos.date;
		} else {
			logMsg("rmp400POM_POSPosterRead() failed\n");
			pomPosterId = NULL;
			mePos->date1 = -1;
		}
	} else
		mePos->date1 = -1;

	configId = &rmp400DataStrId->robot;
	mePos->main.euler.yaw = configId->theta;
	mePos->main.euler.x   = configId->xRob;
	mePos->main.euler.y   = configId->yRob;

	config3dId = &rmp400DataStrId->robot3d;
	mePos->main.euler.pitch = config3dId->pitch;
	mePos->main.euler.roll = config3dId->roll;
	mePos->main.euler.z = config3dId->zRob;

	/* speeds are mostly zero, just update the linear
	   and angular components */
	mePos->VLocal.euler.x   = configId->v;
	mePos->VLocal.euler.yaw = configId->w;

	/* all variances and co-variances remain zero */

	return OK;
}

/*----------------------------------------------------------------------*/

/*
 * Convert joystick values to reference speed
 *
 * A square function is used to get more precision at low speeds.
 *
 */
static void
getJoystickSpeeds(double *v, double *w, double *avmax, double *awmax)
{
	double vgainw = -2, vgainv = -2, againw = -2, againv = -2;
	if (joy->buttons[0]) { vgainw = 0.7; vgainv = 0.4; againw = 0.4; againv = 0.2; }
	if (joy->buttons[1]) { vgainw = 1.0; vgainv = 1.0; againw = 0.5; againv = 0.4; }
	if (joy->buttons[2]) { vgainw = 1.3; vgainv = 2.0; againw = 0.6; againv = 0.6; }
	if (vgainv > -1) {
		*v = -SIGN(joy->axes[1]) * joy->axes[1] * joy->axes[1]
		  * RMP400_JOYSTICK_VGAIN * vgainv;
		*w = -SIGN(joy->axes[0]) * joy->axes[0] * joy->axes[0]
		  * RMP400_JOYSTICK_WGAIN * vgainw;
		if (avmax) *avmax = RMP400_ACCEL_LIN_MAX * againv;
		if (awmax) *awmax = RMP400_ACCEL_ANG_MAX * againw;
	}
	else {
		*v = 0;
		*w = 0;
	}
}
/*------------------------------------------------------------------------
 * Gyro
 *
 * Description: Setup gyroscope
 *
 * Reports:      OK
 *              S_rmp400_GYRO_ERROR
 */

/* rmp400GyroExec  -  codel EXEC of Gyro
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
rmp400GyroExec(RMP400_GYRO_PARAMS *gyroParams, int *report)
{
	GENPOS_CART_STATE *robot = &SDI_F->robot;
	RMP400_GYRO *gyro = &SDI_F->gyro;

	if (gyroParams->mode == gyro->currentMode)
		return ETHER;

	if (gyroParams->mode == RMP400_GYRO_OFF) {
		gyro->currentMode = RMP400_GYRO_OFF;
		return ETHER;
	}

#ifdef GYRO
	if (!gyro->initDone) {
		if (!(gyro->initDone = gyroInitAngle(gyroParams->port))) {
			*report = S_rmp400_GYRO_ERROR;
			gyro->params.mode = RMP400_GYRO_OFF;
			gyro->currentMode = RMP400_GYRO_OFF;
			return ETHER;
		}
	}

	/* read gyro once */
	if (!(gyroReadCorrAngle(&gyro->gyroTheta))) {
		*report = S_rmp400_GYRO_ERROR;
		gyro->params.mode = RMP400_GYRO_OFF;
		gyro->currentMode = RMP400_GYRO_OFF;
		gyro->initDone = FALSE;
		return ETHER;
	}
#if (GYRO==gyroKVH1000Task)
	  if (gyroRate(&gyro->gyroOmega, &gyro->gyroTemperature) != OK) {
		*report = S_rmp400_GYRO_ERROR;
		gyro->params.mode = RMP400_GYRO_OFF;
		gyro->currentMode = RMP400_GYRO_OFF;
		gyro->initDone = FALSE;
		return ETHER;
	  }
#endif

	/* reset gyro offset to match odo */
	gyro->gyroToRobotOffset = robot->theta - gyro->gyroTheta;

	gyro->currentMode = gyro->params.mode;

#else
	printf("rmp400GyroExec: rmp400 module compiled without gyro support. "
	    "Use -DGYRO\n");
	gyro->params.mode = RMP400_GYRO_OFF;
	gyro->currentMode = RMP400_GYRO_OFF;
	gyro->initDone = FALSE;
	*report = S_rmp400_GYRO_ERROR;
#endif

	return ETHER;
}


/*------------------------------------------------------------------------
 * Log
 *
 * Description:
 *
 * Reports:      OK
 *              S_rflex_CANNOT_OPEN_FILE
 *              S_rflex_CANNOT_WRITE_FILE
 */

static FILE *rmp400LogFile;

/* rmp400LogStart  -  codel START of Log
	Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
rmp400LogStart(GEN_STRING128 *fileName, int *report)
{
	if (!(rmp400LogFile = fopen(fileName->name, "a"))) {
		printf ("rmp400LogStart: cannot open odometer file %s\n", fileName->name);
		*report = S_rmp400_CANNOT_OPEN_FILE;
		return ETHER;
	}

	if (!fprintf(rmp400LogFile, "#date\tdate1\tdate2\tmain/yaw\tmain/pitch\tmain/roll\tmain/x\tmain/y\tmain/z\tvlocal/yaw\tvlocal/pitch\tvlocal/roll\tvlocal/x\tvlocal/y\tvlocal/z\n")) {
		printf ("rmp400LogStart: cannot log in %s\n", fileName->name);
		*report = S_rmp400_CANNOT_WRITE_FILE;
		rmp400LogInter(fileName, report);
		return ETHER;
	}

	printf ("rmp400LogStart: log file %s found\n", fileName->name);
	return EXEC;
}

/* rmp400LogExec  -  codel EXEC of Log
	Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
rmp400LogExec(GEN_STRING128 *fileName, int *report)
{
	H2TIMESPEC tv;

	/* get date */
	h2GetTimeSpec(&tv);

	if (!fprintf(rmp400LogFile, "%lf\t%d\t%d\t%2.6g\t%2.6g\t%2.6g\t%2.6g\t%2.6g\t%2.6g\t%2.6g\t%2.6g\t%2.6g\t%2.6g\t%2.6g\t%2.6g\n",
		tv.tv_sec + tv.tv_nsec*1.0E-9, SDI_F->mePos.date1, SDI_F->mePos.date2,
		SDI_F->mePos.main.euler.yaw, SDI_F->mePos.main.euler.pitch, SDI_F->mePos.main.euler.roll,
		SDI_F->mePos.main.euler.x, SDI_F->mePos.main.euler.y, SDI_F->mePos.main.euler.z,
		SDI_F->mePos.VLocal.euler.yaw, SDI_F->mePos.VLocal.euler.pitch, SDI_F->mePos.VLocal.euler.roll,
		SDI_F->mePos.VLocal.euler.x, SDI_F->mePos.VLocal.euler.y, SDI_F->mePos.VLocal.euler.z
		)) {
		printf ("rmp400LogExec: cannot log in %s\n", fileName->name);
		*report = S_rmp400_CANNOT_WRITE_FILE;
		rmp400LogInter(fileName, report);
		return ETHER;
	}

	return EXEC;
}


/* rmp400LogInter  -  codel INTER of Log
	Returns:  INTER ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
rmp400LogInter(GEN_STRING128 *fileName, int *report)
{
	if (rmp400LogFile) {
		fclose(rmp400LogFile);
		rmp400LogFile = NULL;
	}
	return ETHER;
}

/*------------------------------------------------------------------------
 * LogInternalData
 *
 * Description: Log all the raw data provided by robot
 *
 * Reports:     OK
 *              S_rmp400_CANNOT_OPEN_FILE
 *              S_rmp400_CANNOT_WRITE_FILE
 */

static FILE *rmp400LogInternalDataFile;

/* rmp400LogStart  -  codel START of Log
	Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
rmp400LogInternalDataStart(GEN_STRING128 *fileName, int *report)
{
  if (!(rmp400LogInternalDataFile = fopen(fileName->name, "a"))) {
    printf ("rmp400LogInternalDataStart: cannot open internal data file %s\n", fileName->name);
    *report = S_rmp400_CANNOT_OPEN_FILE;
    return ETHER;
  }
  
  if (!fprintf(rmp400LogInternalDataFile,
	       "#date\tdate1\tdate2"
	       "\tgyroAngle"
#if (GYRO==gyroKVH1000Task)
	       "\tgyroRate\tgyroTemperature"
#endif
	       "\tintegrated_yaw1\tyaw_rate1"
	       "\tpitch_angle1\tpitch_rate1"
	       "\troll_angle1\troll_rate1"
	       "\tlw_velocity1\trw_velocity1"
	       "\tintegrated_left_wheel1\tintegrated_right_wheel1"
	       "\tintegrated_fore_aft1\tleft_torque1\tright_torque1"
	       "\tservo_frames1\toperational_mode1"
	       "\tcontroller_gain_schedule1\tui_voltage1"
	       "\tpowerbase_voltage1\tbattery_charge1"
	       "\tvelocity_command1\tturn_command1"
	       "\tintegrated_yaw2\tyaw_rate2"
	       "\tpitch_angle2\tpitch_rate2"
	       "\troll_angle2\troll_rate2"
	       "\tlw_velocity2\trw_velocity2"
	       "\tintegrated_left_wheel2\tintegrated_right_wheel2"
	       "\tintegrated_fore_aft2\tleft_torque2\tright_torque2"
	       "\tservo_frames2\toperational_mode2"
	       "\tcontroller_gain_schedule2\tui_voltage2"
	       "\tpowerbase_voltage2\tbattery_charge2"
	       "\tvelocity_command2\tturn_command2\n")) {
    printf ("rmp400LogStart: cannot log in %s\n", fileName->name);
    *report = S_rmp400_CANNOT_WRITE_FILE;
    rmp400LogInternalDataInter(fileName, report);
    return ETHER;
  }
  printf ("rmp400LogInternalDataStart: log file %s found\n", fileName->name);
  return EXEC;
}

/* rmp400LogInternalDataExec  -  codel EXEC of Log
	Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
rmp400LogInternalDataExec(GEN_STRING128 *fileName, int *report)
{
  H2TIMESPEC tv;
  
  /* get date */
  h2GetTimeSpec(&tv);
  
  if (!fprintf(rmp400LogInternalDataFile, "%lf\t%d\t%d"
	       "\t%.6g"
#if (GYRO==gyroKVH1000Task)
	       "\t%.6g\t%.6g"
#endif
	       "\t%.6g\t%.6g"
	       "\t%.6g\t%.6g"
	       "\t%.6g\t%.6g"
	       "\t%.6g\t%.6g"
	       "\t%.6g\t%.6g"
	       "\t%.6g\t%.6g\t%.6g"
	       "\t%d\t%d"
	       "\t%d\t%.6g"
	       "\t%.6g\t%.6g"
	       "\t%.6g\t%.6g"
	       "\t%.6g\t%.6g"
	       "\t%.6g\t%.6g"
	       "\t%.6g\t%.6g"
	       "\t%.6g\t%.6g"
	       "\t%.6g\t%.6g"
	       "\t%.6g\t%.6g\t%.6g"
	       "\t%d\t%d"
	       "\t%d\t%.6g"
	       "\t%.6g\t%.6g"
	       "\t%.6g\t%.6g"
	       "\n",
				/* timestamps */
	       tv.tv_sec + tv.tv_nsec*1.0E-9, 
	       SDI_F->mePos.date1, SDI_F->mePos.date2,
				/* the KVH gyro */
	       SDI_F->gyro.gyroTheta,
#if (GYRO==gyroKVH1000Task)
	       SDI_F->gyro.gyroOmega, SDI_F->gyro.gyroTemperature,
#endif
				/* the robot first axle */
	       SDI_F->status.rs_data[0].integrated_yaw,
	       SDI_F->status.rs_data[0].yaw_rate,
	       SDI_F->status.rs_data[0].pitch_angle,
	       SDI_F->status.rs_data[0].pitch_rate,
	       SDI_F->status.rs_data[0].roll_angle,
	       SDI_F->status.rs_data[0].roll_rate,
	       SDI_F->status.rs_data[0].lw_velocity,
	       SDI_F->status.rs_data[0].rw_velocity,
	       SDI_F->status.rs_data[0].integrated_left_wheel,
	       SDI_F->status.rs_data[0].integrated_right_wheel,
	       SDI_F->status.rs_data[0].integrated_fore_aft,
	       SDI_F->status.rs_data[0].left_torque,
	       SDI_F->status.rs_data[0].right_torque,
	       SDI_F->status.rs_data[0].servo_frames,
	       SDI_F->status.rs_data[0].operational_mode,
	       SDI_F->status.rs_data[0].controller_gain_schedule,
	       SDI_F->status.rs_data[0].ui_voltage,
	       SDI_F->status.rs_data[0].powerbase_voltage,
	       SDI_F->status.rs_data[0].battery_charge,
	       SDI_F->status.rs_data[0].velocity_command,
	       SDI_F->status.rs_data[0].turn_command,
				/* the robot second axle */
	       SDI_F->status.rs_data[1].integrated_yaw,
	       SDI_F->status.rs_data[1].yaw_rate,
	       SDI_F->status.rs_data[1].pitch_angle,
	       SDI_F->status.rs_data[1].pitch_rate,
	       SDI_F->status.rs_data[1].roll_angle,
	       SDI_F->status.rs_data[1].roll_rate,
	       SDI_F->status.rs_data[1].lw_velocity,
	       SDI_F->status.rs_data[1].rw_velocity,
	       SDI_F->status.rs_data[1].integrated_left_wheel,
	       SDI_F->status.rs_data[1].integrated_right_wheel,
	       SDI_F->status.rs_data[1].integrated_fore_aft,
	       SDI_F->status.rs_data[1].left_torque,
	       SDI_F->status.rs_data[1].right_torque,
	       SDI_F->status.rs_data[1].servo_frames,
	       SDI_F->status.rs_data[1].operational_mode,
	       SDI_F->status.rs_data[1].controller_gain_schedule,
	       SDI_F->status.rs_data[1].ui_voltage,
	       SDI_F->status.rs_data[1].powerbase_voltage,
	       SDI_F->status.rs_data[1].battery_charge,
	       SDI_F->status.rs_data[1].velocity_command,
	       SDI_F->status.rs_data[1].turn_command))
    {
      printf ("rmp400LogInternalDataExec: cannot log in %s\n", 
	      fileName->name);
      *report = S_rmp400_CANNOT_WRITE_FILE;
      rmp400LogInter(fileName, report);
      return ETHER;
    }
  
  return EXEC;
}

/* rmp400LogInternalDataInter  -  codel INTER of LogInternalData
	Returns:  INTER ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
rmp400LogInternalDataInter(GEN_STRING128 *fileName, int *report)
{
  if (rmp400LogInternalDataFile) {
    fclose(rmp400LogInternalDataFile);
    rmp400LogInternalDataFile = NULL;
  }
  return ETHER;
}



#ifdef WITH_FELIB
/*------------------------------------------------------------------------
 * ResetEmergencyStop
 *
 * Description: 
 *
 * Reports:      OK
 *              S_rmp400_EMERGENCY_STOP
 */

/* resMain  -  codel EXEC of ResetEmergencyStop
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
resMain(int *report)
{
	RMP400_STATUS_STR *status = &SDI_F->status;
	int i;

	if ((fe_pins(fe) & (FE_PAUSE|FE_STOP)) != 0) {
		*report = S_rmp400_EMERGENCY_STOP;
		return END;
	}
	status->rs_mode = RMP400_IDLE;

	return ETHER;
}
#endif
